﻿// by ChatGPT 4o:
// Query: I have an instance of some data structure in C# - e.g. Queue<int>.
//        I can use Monitor.Wait and Pulse methods to communicate a state of a condition related to the data structure.
//        How should I implement if I want to have a second independent condition on the same data structure instance?
// Reply: When you need to implement multiple independent conditions on a data structure like Queue<int> in C# using Monitor.Wait and Monitor.Pulse,
//        you must carefully manage the synchronization to ensure that each condition can be independently waited on and signaled.
//        Here’s a step-by-step approach to implementing this:
//        (only the code is copy-pasted here + MinItems was made a public constant instead of private field + marked comments added)

class QueueWithConditions {
	private Queue<int> _queue = new Queue<int>();

	private object _lockObj = new object();
	private object _notEmptyLock = new object();
	private object _hasMinItemsLock = new object();

	public const int MinItems = 5;

	// Enqueue item and signal conditions if necessary
	public void Enqueue(int item) {
		lock (_lockObj) {
			_queue.Enqueue(item);

			// Signal that the queue is not empty
			lock (_notEmptyLock) {
				Monitor.Pulse(_notEmptyLock);
			}

			// Signal that the queue has more than _minItems items
			if (_queue.Count >= MinItems) {
				lock (_hasMinItemsLock) {
					Monitor.Pulse(_hasMinItemsLock);
				}
			}
		}
	}

	// Dequeue item and signal conditions if necessary
	public int Dequeue() {
		int item;
		lock (_lockObj) {
			item = _queue.Dequeue();
		}

		return item;
	}

	// Wait for the queue to be not empty
	public void WaitUntilNotEmpty() {
		lock (_notEmptyLock) {
			while (_queue.Count == 0) {
				Monitor.Wait(_notEmptyLock);
			}
		}
	}

	// Wait for the queue to have at least _minItems items
	public void WaitUntilHasMinItems() {
		lock (_hasMinItemsLock) {
			while (_queue.Count < MinItems) {
				Monitor.Wait(_hasMinItemsLock);
			}
		}
	}
}

// Example usage
class Program {
	static void Main() {
		QueueWithConditions queue = new QueueWithConditions();

		// Example producer
		Thread producer = new Thread(() => {
			for (int i = 0; i < 10; i++) {
				queue.Enqueue(i);
				Console.WriteLine("Enqueued: " + i);
				Thread.Sleep(100); // Simulate work
			}
		});

		// Example consumer
		Thread consumer = new Thread(() => {
			while (true) {
				queue.WaitUntilNotEmpty();
				int item = queue.Dequeue();	// Comment by Sonic: BUG: Race condition here !!!

				Console.WriteLine("Dequeued: " + item);
				Thread.Sleep(500); // Simulate work
			}
		});

		// Example monitor for min items condition
		Thread minItemsMonitor = new Thread(() => {
			while (true) {
				queue.WaitUntilHasMinItems();
				
				Console.WriteLine($"Queue has at least {QueueWithConditions.MinItems} items");
				Thread.Sleep(100); // Simulate work
			}
		});

		producer.Start();
		consumer.Start();
		minItemsMonitor.Start();

		producer.Join();
		consumer.Join();    // Comment by Sonic: BUG: Actually never ends !
		minItemsMonitor.Join(); // Comment by Sonic: BUG: Actually never ends !
	}
}
